#include <iostream>
#include <cmath>
#include <random>
/*
  On utilise la bibliotheque random, presente seulement depuis C++ 11.
  Pour compiler ce programme avec g++, ajouter l'option -std=c++11.

  Sinon, remplacer l'utilisation de mt19937 par l'utilisation de rand.
*/ 


/*
  Rappel : si X est une variable aleatoire de fonction de
  repartition F et que G est l'inverse generalise de F
  (c'est-a-dire G(t) = inf {x, F(x) >= t} ), alors la variable
  G(U), ou U est uniforme sur [0,1] suit la meme loi que X.
 */


const int N = 100000; // Nombre de simulation dans les calculs par Monte Carlo
const double p = 0.3; // Parametre des lois de Bernoulli


////////////////////////////////////////

// Definition d'un certain nombre de fonctions renvoyant des nombres
// aleatoire de loi determinee.


double unif(std::mt19937 &generateur)
// Simulation d'une loi uniforme sur [0,1]
{
  std::uniform_real_distribution<double> U;
  return U(generateur);
}

double bernoulli(std::mt19937 &generateur)
{
  std::bernoulli_distribution X(p);
  return X(generateur);
}

double normale(std::mt19937 &generateur)
{
  std::normal_distribution<double> G(0,1);
  return G(generateur);
}

double expo(std::mt19937 &generateur)
{
  std::exponential_distribution<double> X(1);
  return X(generateur);
}

double cauchy(std::mt19937 &generateur)
{
  std::cauchy_distribution<double> X(1);
  return X(generateur);
}

double densite_expl(std::mt19937 &generateur)
{
  // Methode de la fonction de repartition inverse : si U est
  // uniforme sur [0,1], alors U^-1/2 a pour densite 2x^-3 sur [1,+oo[.
  return 1/sqrt(unif(generateur));
}

////////////////////////////////////////////////////////////

void affiche_intervalle(double somme, double somme_carres)
/*
  Affiche un intervalle de confiance a 95% sur la moyenne de X
  a partir de :
  - la somme des x_i (somme)
  - la somme des x_i^2 (somme_carres)
  La taille de l'echantillon considere est N.
*/
{
  double moyenne, variance;
  moyenne = somme/N;
  variance = somme_carres/N - moyenne*moyenne;

  double largeur = 1.96*sqrt(variance/N);

  std::cout << "Intervalle de confiance a 95% : [ " << moyenne-largeur << " , " << moyenne+largeur << " ], centre en " << moyenne << "\n";
}


// La fonction suivante calcule un intervalle de confiance a partir d'un genrateur aleatoire et
// d'une fonction qui renvoie des reels quand on lui fourni un generateur aleatoire.
void calcul_IC(double (*f)(std::mt19937 &gen), std::mt19937 &gen)
{
  double somme, somme_carres, X;
  somme = 0;
  somme_carres = 0;
  for (int i=0; i<N; i++)
	{
	  X = f(gen);
	  somme += X;
	  somme_carres += X*X;
	}
  affiche_intervalle(somme, somme_carres);
}


/////////////////////////////////////////////////////////

// On utilise la fonction calcul_IC pour tous les exemples precedents.

int main(void)
{
  std::mt19937 generateur(12345); // Le generateur que l'on va fournir en argument a toute fonction produisant de l'alea.
  
  std::cout << "Moyenne d'une loi uniforme sur [0,1] :\n";
  calcul_IC(unif, generateur);
  std::cout << "Moyenne theorique : 0.5\n\n";
  
  std::cout << "Moyenne d'une loi de Bernoulli de parametre 0.3 :\n";
  calcul_IC(bernoulli, generateur);
  std::cout << "Moyenne theorique : 0.3\n\n";
  
  std::cout << "Moyenne d'une loi N(0,1) :\n";
  calcul_IC(normale, generateur);
  std::cout << "Moyenne theorique : 0\n\n";
  
  std::cout << "Moyenne d'une loi exponentielle de parametre 1 :\n";
  calcul_IC(expo, generateur);
  std::cout << "Moyenne theorique : 1\n\n";
  
  std::cout << "Moyenne d'une loi de Cauchy :\n";
  calcul_IC(cauchy, generateur);
  std::cout << "Pas de moyenne theorique : la loi de Cauchy n'est pas integrable\n\n";
  
  std::cout << "Moyenne d'une loi de densite 2x^-3 sur [1,+oo[ :\n";
  calcul_IC(densite_expl, generateur);
  std::cout << "Moyenne theorique : 2. Noter que cette loi n'est pas de carre integrable, et que l'intervalle de confiance n'a donc pas de justification theorique.\n\n";
  
  return 0;
}

